//! Run QEMU.

use std::{
    fs::{copy, File},
    path::{Path, PathBuf},
    process::Command,
};

use bytesize::MIB;
use log::{debug, error, trace};
use tempfile::tempdir_in;

use crate::{
    config2::Config,
    project::Project,
    vdrive::{VirtualDrive, VirtualDriveError},
};

const OVMF_FD: &str = "/usr/share/ovmf/OVMF.fd";

pub const EXECUTOR_DRIVE: &str = "/dev/vdb";
pub const SOURCE_DRIVE: &str = "/dev/vdc";
pub const ARTIFACT_DRIVE: &str = "/dev/vdd";
pub const CACHE_DRIVE: &str = "/dev/vde";
pub const DEPS_DRIVE: &str = "/dev/vdf";

pub const WORKSPACE_DIR: &str = "/workspace";
pub const SOURCE_DIR: &str = "/workspace/src";
pub const DEPS_DIR: &str = "/workspace/deps";
pub const CACHE_DIR: &str = "/workspace/cache";
pub const ARTIFACTS_DIR: &str = "/workspace/artifacts";

#[derive(Default)]
pub struct QemuRunner<'a> {
    config: Option<&'a Config>,
    project: Option<&'a Project>,
    executor: Option<&'a VirtualDrive>,
    source: Option<&'a VirtualDrive>,
    dependencies: Option<&'a VirtualDrive>,
    cache: Option<&'a VirtualDrive>,
    artifacts: Option<&'a VirtualDrive>,
    log: Option<&'a Path>,
}

impl<'a> QemuRunner<'a> {
    pub fn config(mut self, value: &'a Config) -> Self {
        self.config = Some(value);
        self
    }

    pub fn project(mut self, value: &'a Project) -> Self {
        self.project = Some(value);
        self
    }

    pub fn excutor(mut self, value: &'a VirtualDrive) -> Self {
        self.executor = Some(value);
        self
    }

    pub fn source(mut self, value: &'a VirtualDrive) -> Self {
        self.source = Some(value);
        self
    }

    pub fn dependencies(mut self, value: &'a VirtualDrive) -> Self {
        self.dependencies = Some(value);
        self
    }

    pub fn cache(mut self, value: &'a VirtualDrive) -> Self {
        self.cache = Some(value);
        self
    }

    pub fn artifacts(mut self, value: &'a VirtualDrive) -> Self {
        self.artifacts = Some(value);
        self
    }

    pub fn log(mut self, value: &'a Path) -> Self {
        self.log = Some(value);
        self
    }

    pub fn run(&self) -> Result<(), QemuError> {
        let config = self.config.ok_or(QemuError::Missing("config"))?;
        let project = self.project.ok_or(QemuError::Missing("project"))?;
        let executor_drive = self.executor.ok_or(QemuError::Missing("executor_drive"))?;
        let source = self.source.ok_or(QemuError::Missing("source"))?;
        let dependencies = self
            .dependencies
            .ok_or(QemuError::Missing("dependencies"))?;
        let cache = self.cache.ok_or(QemuError::Missing("cache"))?;
        let artifacts = self.artifacts.ok_or(QemuError::Missing("artifacts"))?;
        let log = self.log.ok_or(QemuError::Missing("log"))?;

        let qemu = Qemu {
            image: project.image().into(),
            tmpdir: config.tmpdir().into(),
            log: log.into(),
            executor: Some(executor_drive),
            source: Some(source),
            artifact: Some(artifacts),
            dependencies: Some(dependencies),
            cache: Some(cache),
            cpus: config.cpus(),
            memory: config.memory().as_u64(),
        };
        let exit = qemu.run()?;
        debug!("QEMU exit code {}", exit);

        if exit != 0 {
            return Err(QemuError::QemuFailed(exit));
        }

        Ok(())
    }
}

/// A QEMU runner.
#[derive(Debug)]
struct Qemu<'a> {
    image: PathBuf,
    tmpdir: PathBuf,
    log: PathBuf,
    executor: Option<&'a VirtualDrive>,
    source: Option<&'a VirtualDrive>,
    artifact: Option<&'a VirtualDrive>,
    dependencies: Option<&'a VirtualDrive>,
    cache: Option<&'a VirtualDrive>,
    cpus: usize,
    memory: u64,
}

impl<'a> Qemu<'a> {
    /// Run QEMU in the specified way.
    fn run(&self) -> Result<i32, QemuError> {
        debug!("run QEMU");
        trace!("{:#?}", self);
        let tmp = tempdir_in(&self.tmpdir).map_err(QemuError::TempDir)?;

        let image = tmp.path().join("vm.qcow2");
        let vars = tmp.path().join("vars.fd");
        let console_log = tmp.path().join("console.log");

        debug!("create copy-on-write image and UEFI vars file");
        create_cow_image(&self.image, &image)?;
        copy(OVMF_FD, &vars).map_err(|e| QemuError::Copy(OVMF_FD.into(), e))?;

        let source = self.source.unwrap();
        let artifact_drive = self.artifact.unwrap();
        let deps_drive = self.dependencies.unwrap();
        let cache_drive = self.cache.unwrap();
        let executor_drive = self.executor.unwrap();

        debug!("set console log file");
        Self::create_file(&console_log)?;

        debug!("set build log file");
        let build_log = Self::create_file(&self.log)?;

        let cpus = format!("cpus={}", self.cpus);
        let memory = format!("{}", self.memory / MIB);
        let args = QemuArgs::default()
            .with_valued_arg("-m", &memory)
            .with_valued_arg("-smp", &cpus)
            .with_valued_arg("-display", "none")
            .with_valued_arg("-serial", &format!("file:/{}", console_log.display())) // ttyS0
            .with_valued_arg("-serial", &format!("file:{}", build_log.display())) // ttyS1
            .with_ipflash(0, OVMF_FD, true)
            .with_ipflash(1, vars.to_str().unwrap(), false)
            .with_qcow2(image.to_str().unwrap())
            .with_raw(executor_drive.filename(), true)
            .with_raw(source.filename(), true)
            .with_raw(artifact_drive.filename(), false)
            .with_raw(cache_drive.filename(), false)
            .with_raw(deps_drive.filename(), true)
            .with_arg("-nodefaults");

        debug!("spawn QEMU");
        let child = Command::new("kvm")
            .args(args.iter())
            .spawn()
            .map_err(QemuError::Invoke)?;

        debug!("wait for QEMU to finish");
        let output = child.wait_with_output().map_err(QemuError::Run)?;
        if output.status.success() {
            debug!("QEMU finished OK");
            let log = Self::build_log(&build_log)?;
            let exit = Self::exit_code(&log)?;
            Ok(exit)
        } else {
            debug!("QEMU failed");
            let exit = output.status.code().unwrap_or(255);
            let out = String::from_utf8_lossy(&output.stdout).to_string();
            let err = String::from_utf8_lossy(&output.stderr).to_string();
            let _log = Self::build_log(&build_log)?;
            let log = std::fs::read(&console_log).map_err(QemuError::TemporaryLog)?;
            let log = String::from_utf8_lossy(&log).to_string();
            error!(
                "QEMU failed: exit={}\nstdout: {:?}\nstderr: {:?}\nVM console: {:?}",
                exit, out, err, log
            );
            Ok(exit)
        }
    }

    fn create_file(filename: &Path) -> Result<PathBuf, QemuError> {
        File::create(filename).map_err(|e| QemuError::Log(filename.into(), e))?;
        Ok(filename.into())
    }

    fn build_log(filename: &Path) -> Result<String, QemuError> {
        const BEGIN: &str = "====================== BEGIN ======================";

        let log = std::fs::read(filename).map_err(|e| QemuError::ReadLog(filename.into(), e))?;
        let log = String::from_utf8_lossy(&log);
        trace!("build log from QEMU:\n{log}");

        if let Some((_, log)) = log.split_once(BEGIN) {
            return Ok(log.into());
        }

        Ok("".into())
    }

    fn exit_code(log: &str) -> Result<i32, QemuError> {
        const EXIT: &str = "\nEXIT CODE: ";
        if let Some((_, rest)) = log.split_once(EXIT) {
            trace!("log has {:?}", EXIT);
            if let Some((exit, _)) = rest.split_once('\n') {
                let exit = exit.trim();
                trace!("log has newline after exit: {:?}", exit);
                if let Ok(exit) = exit.parse::<i32>() {
                    trace!("log exit coded parses ok: {}", exit);
                    return Ok(exit);
                }
                debug!("log exit does not parse");
            }
        }
        Err(QemuError::NoExit)
    }
}

fn create_cow_image(backing_file: &Path, new_file: &Path) -> Result<(), QemuError> {
    let output = Command::new("qemu-img")
        .arg("create")
        .arg("-b")
        .arg(backing_file)
        .args(["-F", "qcow2"])
        .args(["-f", "qcow2"])
        .arg(new_file)
        .output()
        .map_err(QemuError::Run)?;
    if !output.status.success() {
        let exit = output.status.code().unwrap_or(255);
        let out = String::from_utf8_lossy(&output.stdout).to_string();
        let err = String::from_utf8_lossy(&output.stderr).to_string();
        error!(
            "qemu-img failed: exit={}\nstdout: {:?}\nstderr: {:?}",
            exit, out, err,
        );
        Err(QemuError::QemuImg(exit, err))
    } else {
        Ok(())
    }
}

#[derive(Debug, Default)]
struct QemuArgs {
    args: Vec<String>,
}

impl QemuArgs {
    fn with_arg(mut self, arg: &str) -> Self {
        self.args.push(arg.into());
        self
    }

    fn with_valued_arg(mut self, arg: &str, value: &str) -> Self {
        self.args.push(arg.into());
        self.args.push(value.into());
        self
    }

    fn with_ipflash(mut self, unit: usize, path: &str, readonly: bool) -> Self {
        self.args.push("-drive".into());
        self.args.push(format!(
            "if=pflash,format=raw,unit={},file={}{}",
            unit,
            path,
            if readonly { ",readonly=on" } else { "" },
        ));
        self
    }

    fn with_qcow2(mut self, path: &str) -> Self {
        self.args.push("-drive".into());
        self.args
            .push(format!("format=qcow2,if=virtio,file={}", path));
        self
    }

    fn with_raw(mut self, path: &Path, readonly: bool) -> Self {
        self.args.push("-drive".into());
        self.args.push(format!(
            "format=raw,if=virtio,file={}{}",
            path.display(),
            if readonly { ",readonly=on" } else { "" },
        ));
        self
    }

    fn iter(&self) -> impl Iterator<Item = &str> {
        self.args.iter().map(|s| s.as_str())
    }
}

/// Possible errors from running Qemu.
#[allow(missing_docs)]
#[derive(Debug, thiserror::Error)]
pub enum QemuError {
    #[error("missing field in QemuBuilder: {0}")]
    Missing(&'static str),

    #[error("failed to invoke QEMU")]
    Invoke(#[source] std::io::Error),

    #[error("QEMU process failed")]
    Run(#[source] std::io::Error),

    #[error("qemu-img failed: exit code {0}: stderr:\n{1}")]
    QemuImg(i32, String),

    #[error("failed to create a temporary directory")]
    TempDir(#[source] std::io::Error),

    #[error("failed to copy to temporary directory: {0}")]
    Copy(PathBuf, #[source] std::io::Error),

    #[error("failed to read log file {0}")]
    ReadLog(PathBuf, #[source] std::io::Error),

    #[error("failed to open log file for writing")]
    Log(PathBuf, #[source] std::io::Error),

    #[error("failed to create a tar archive from {0}")]
    Tar(PathBuf, #[source] VirtualDriveError),

    #[error("failed to extract cache drive to {0}")]
    ExtractCache(PathBuf, #[source] VirtualDriveError),

    #[error("failed to read temporary file for logging")]
    TemporaryLog(#[source] std::io::Error),

    #[error("build log lacks exit code of run")]
    NoExit,

    #[error("failed to get length of file {0}")]
    Metadata(PathBuf, #[source] std::io::Error),

    #[error("failed to set length of file to {0}: {1}")]
    SetLen(u64, PathBuf, #[source] std::io::Error),

    #[error("failed to run actions in QEMU")]
    QemuFailed(i32),
}
